#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
    qa-bughandler
    ~~~~~~~~~~~~~

    :copyright: (c) 2012 Collabora Ltd.
    :author: Rafael G. Martins <rafael.martins@collabora.com>
    :license: LGPL-2.1+

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301 USA
"""

from argparse import ArgumentParser
from ConfigParser import ConfigParser
from datetime import datetime
from traceback import format_exc
from urlparse import urlparse, urlunparse
from xmlrpclib import ServerProxy, ProtocolError
import logging
import os
import sys

__version__ = '0.1'

LOG_FORMATTER = "[%(asctime)s] %(name)s.%(levelname)s: %(message)s"
LOG_DATEFORMAT = "%Y-%m-%d %H:%M:%S %Z"


class Config(object):
    """This class reads some settings from a configuration file. It depends on
    the file name that comes from the Cli object.
    """

    def __init__(self, filename):
        self.filename = filename
        self.parser = ConfigParser()
        if not self.parser.read(self.filename):
            raise RuntimeError('Configuration file not found: %s' %
                               self.filename)
        if not self.parser.has_section('settings'):
            raise RuntimeError('[settings] section not found.')

    def get(self, option, default=None):
        """Reads a setting from the [settings] section of the configuration
        file. If not found, the value of the 'default' parameter will be
        returned.
        """
        if not self.parser.has_option('settings', option):
            return default
        return self.parser.get('settings', option)

    def __getattr__(self, option):
        """Anything from the [settings] section will be loaded automatically
        as an attribute of this class. If a setting isn't found, this method
        will raise RuntimeError.
        """
        if not self.parser.has_option('settings', option):
            raise RuntimeError('[settings].%s option not found.' % option)
        return self.parser.get('settings', option)


class Cli(object):
    """This class reads arguments from the command-line interface, using the
    argparse module.
    """

    def __init__(self):
        self.parser = ArgumentParser()

        # declaration of all the argparse arguments
        self.parser.add_argument('--status', '-s', dest='status',
                                 metavar='STATUS', help='modify bug status')
        self.parser.add_argument('--resolution', '-r', dest='resolution',
                                 metavar='RESOLUTION',
                                 help='modify bug resolution')
        self.parser.add_argument('--image', '-i', dest='image',
                                 metavar='IMAGE', help='the name of the '
                                 'image used')
        self.parser.add_argument('--package', '-p', dest='package',
                                 metavar='PACKAGE', help='the name of the '
                                 'package, with the version, that fixes the '
                                 'bug')
        self.parser.add_argument('--field', '-f', dest='fields',
                                 action='append', metavar='fieldname=value',
                                 help='modify the given fields.')
        self.parser.add_argument('--config-file', '-c', dest='config_file',
                                 metavar='FILE', help='configuration file, '
                                 'defaults to `~/.qa-bughandler.conf\'',
                                 default=os.path.expanduser(
                                     '~/.qa-bughandler.conf'))
        self.parser.add_argument('--log-level', dest='log_level',
                                 choices=['CRITICAL', 'ERROR', 'WARNING',
                                          'INFO', 'DEBUG'],
                                 default='INFO', help='logging level.')
        self.parser.add_argument('--log-file', dest='log_file', metavar='FILE',
                                 help='send logging data to FILE, instead '
                                 'of stderr.')
        self.parser.add_argument('--comment', dest='comment', metavar='COMMENT',
                                 help='Extra text to add to the update comment')
        self.parser.add_argument('--comment-only', dest='comment_only',
                                 action="store_true",
                                 help='Extra text to add to the update comment')
        self.parser.add_argument('bug_id', help='bug identifier')

        # run the parser through argv (implicit)
        self.args = self.parser.parse_args()


class Settings(object):
    """This class is just a common endpoint for Cli and Config. It also
    provides logging capabilities, that depends on the settings that were
    collected from Cli.
    """

    def __init__(self):
        # load cli settings
        self.cli = Cli()

        # setup logging
        self.logger = logging.getLogger('qa-bughandler')
        self.logger.setLevel(logging.getLevelName(self.cli.args.log_level))
        handler = logging.StreamHandler(sys.stderr)
        if self.cli.args.log_file is not None:
            handler = logging.FileHandler(self.cli.args.log_file, 'a', 'utf-8')
        handler.setFormatter(logging.Formatter(LOG_FORMATTER, LOG_DATEFORMAT))
        self.logger.addHandler(handler)
        self.logger.info('Starting qa-bughandler.py. Version: %s', __version__)

        # load config settings
        self.config = Config(self.cli.args.config_file)


class Bugzilla(object):
    """This class is a XML-RPC client, that handles our calls to the Bugzilla
    API methods.
    """

    def __init__(self, url, username, password, bug_id, logger, verbose=False):
        self.url = url
        self.bug_id = bug_id
        self.logger = logger
        self.verbose = verbose
        self.username = username
        self.password = password
        self.handler = ServerProxy(self.url, verbose=self.verbose)
        self.http_auth = False

    def login(self):
        """This method implements the HTTP login, adding the user name and the
        password to the url, while keeping the url property of the class still
        safe to be displayed/logged.

        """
        try:
                self.handler.Bugzilla.version({})
        except ProtocolError as e:
                if e.errcode != 401:
                    raise e
                self.http_auth = True

        if not self.http_auth:
                return

        self.logger.debug('Login to bugzilla using http auth: %s', self.url)

        # split and unsplit the url with the username and password
        oldurl = urlparse(self.url)
        newurl = (oldurl.scheme,
                  '%s:%s@%s' % (self.username, self.password, oldurl.netloc),
                  oldurl.path, oldurl.params, oldurl.query, oldurl.fragment)

        # rebuild the handler with the new url
        self.handler = ServerProxy(urlunparse(newurl), verbose=self.verbose)

    def add_credentials (self, params):
        if not self.http_auth:
            params.update ({ 'Bugzilla_login': self.username,
                         'Bugzilla_password': self.password })

    def get(self):
        """This method returns data from #bug_id bug."""
        params = {'ids': [self.bug_id]}
        self.add_credentials (params)
        self.logger.debug('XML-RPC call: Bug.get, params: %r', params)
        rv = self.handler.Bug.get(params)
        self.logger.debug('XML-RPC response: %r', rv)

        # 'bugs' is the dict element with a list of bugs data. We just provided
        # one bug_id, then we just have one element in the list.
        return rv['bugs'][0]

    def update(self, **fields):
        """This method updates a bug with, changing the required fields."""
        params = {'ids': [self.bug_id]}
        params.update(fields)
        self.add_credentials (params)
        self.logger.debug('XML-RPC call: Bug.update, params: %r', params)
        rv = self.handler.Bug.update(params)
        self.logger.debug('XML-RPC response: %r', rv)
        return rv

    def add_comment(self, comment):
        """This method adds a comment to the required bug."""
        params = {'id': self.bug_id, 'comment': comment}
        self.add_credentials (params)
        self.logger.debug('XML-RPC call: Bug.add_comment, params: %r', params)
        rv = self.handler.Bug.add_comment(params)
        self.logger.debug('XML-RPC response: %r', rv)
        return rv

def compare_status (a, b):
    # Order
    order = [
        "UNKNOWN_STATUS",
        "FIX_PROPOSED",
        "FIX_SUBMITTED",
        "RESOLVED",
        "VERIFIED",
        "CLOSED" ]
    try:
        a_i = order.index(a)
    except ValueError:
        a_i = 0
    try:
        b_i = order.index(b)
    except ValueError:
        b_i = 0

    return b_i - a_i

def do_main(settings):
    fields = {}

    bugz = Bugzilla(settings.config.url, settings.config.username,
                    settings.config.password, settings.cli.args.bug_id,
                    settings.logger)

    # we need to login before any request
    bugz.login()

    # we should get the bug info to validate the fields, and handle special
    # fields properly
    bug = bugz.get()

    # handle image
    if settings.cli.args.image is not None:

        # image field name comes from the config file, or defaults to 'image'
        field_name = settings.config.get('field_image', 'image')
        fields[field_name] = settings.cli.args.image

    # handle package
    if settings.cli.args.package is not None:

        # package field name comes from the config file, or defaults to
        # 'package'
        field_name = settings.config.get('field_package', 'package')
        # Only update the package field if it's currently empty
        if bug[field_name] == '':
            fields[field_name] = settings.cli.args.package

    # Only update the status if it moved us forward towards CLOSED and only
    # update resolution when also doing a status update
    if settings.cli.args.status is not None:
        if compare_status(bug["status"], settings.cli.args.status) >= 0:
            fields['status'] = settings.cli.args.status

            # only handle resolution if a status is given
            if settings.cli.args.resolution is not None:
                fields['resolution'] = settings.cli.args.resolution

    # handle additional fields
    if settings.cli.args.fields is not None:

        # 'fields' is a list of 'key=value' strings
        for field in settings.cli.args.fields:

            # invalid format, ignore
            if '=' not in field:
                continue

            # everything before the first '=' is the key
            key, value = field.split('=', 1)

            # key and value shouldn't have empty chars around
            fields[key.strip()] = value.strip()

    # 'image' is a special field. it is a list of image names, and we need
    # to just append our current image to it, if needed.
    field_image = settings.config.get('field_image', 'image')
    if field_image in fields:  # we don't need to touch it if no image was
                               # provided :)
        if field_image not in bug:  # field not present in our bz setup.
            settings.logger.warning('Custom field not found: %s', field_image)
        images = [i.strip() for i in bug[field_image].split(',')]
        if fields[field_image] in images:
            fields.pop(field_image)
        else:
            images.append(fields[field_image])
            fields[field_image] = ', '.join(images)


    # We will force all the time/data to be in UTC, to make the things
    # easier to handle
    now = datetime.utcnow()

    # We will add a comment to the bug. For this we need a template,
    # provided in the configuration file. If not found, use a sane default.
    template = settings.config.get('comment_template',
                                   'Ticket updated by {script}, '
                                   '{date} {time} {timezone}:')


    # Only comment handling left
    if fields == {} and not settings.cli.args.comment_only:
        settings.logger.info('Nothing to do')
        return

    # Replace the template variables.
    comment = template.format(script='qa-bughandler',
                              date=now.strftime('%Y-%m-%d'),
                              time=now.strftime('%H:%M:%S'),
                              timezone='UTC')
    comment = comment.strip() + '\n\n'

    if settings.cli.args.comment:
        comment = comment + settings.cli.args.comment + '\n\n'

    # validate fields and add them to the comment.
    for field in fields:
        if field not in bug:
            settings.logger.warning('Field not found: %s', field)
        comment += '* %s: %s\n' % (field, fields[field])

    # update the bug fields & comment in one go
    fields.update ({ "comment": { "body": comment }})
    bugz.update(**fields)

if __name__ == '__main__':
    settings = Settings()
    try:
        do_main(settings)
    except Exception, e:
        settings.logger.critical('%s: %s', e.__class__.__name__, e)
        settings.logger.critical(format_exc())
        sys.exit(-1)
